@@ -1,5 +1,5 @@
 /* Generic implementation of the RESHAPE intrinsic
-   Copyright 2002, 2006, 2007 Free Software Foundation, Inc.
+   Copyright 2002, 2006, 2007, 2009 Free Software Foundation, Inc.
    Contributed by Paul Brook <paul@nowt.org>
 
 This file is part of the GNU Fortran 95 runtime library (libgfortran).
@@ -7,26 +7,21 @@ This file is part of the GNU Fortran 95 runtime library (libgfortran).
 Libgfortran is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public
 License as published by the Free Software Foundation; either
-version 2 of the License, or (at your option) any later version.
-
-In addition to the permissions in the GNU General Public License, the
-Free Software Foundation gives you unlimited permission to link the
-compiled version of this file into combinations with other programs,
-and to distribute those combinations without any restriction coming
-from the use of this file.  (The General Public License restrictions
-do apply in other respects; for example, they cover modification of
-the file, and distribution when not linked into a combine
-executable.)
+version 3 of the License, or (at your option) any later version.
 
 Ligbfortran is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
-You should have received a copy of the GNU General Public
-License along with libgfortran; see the file COPYING.  If not,
-write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-Boston, MA 02110-1301, USA.  */
+Under Section 7 of GPL version 3, you are granted additional
+permissions described in the GCC Runtime Library Exception, version
+3.1, as published by the Free Software Foundation.
+
+You should have received a copy of the GNU General Public License and
+a copy of the GCC Runtime Library Exception along with this program;
+see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
+<http://www.gnu.org/licenses/>.  */
 
 #include "libgfortran.h"
 #include <stdlib.h>
@@ -49,7 +44,7 @@ reshape_internal (parray *ret, parray *source, shape_type *shape,
   index_type rsize;
   index_type rs;
   index_type rex;
-  char *rptr;
+  char * restrict rptr;
   /* s.* indicates the source array.  */
   index_type scount[GFC_MAX_DIMENSIONS];
   index_type sextent[GFC_MAX_DIMENSIONS];
@@ -69,71 +64,43 @@ reshape_internal (parray *ret, parray *source, shape_type *shape,
   const char *src;
   int n;
   int dim;
-  int sempty, pempty;
+  int sempty, pempty, shape_empty;
+  index_type shape_data[GFC_MAX_DIMENSIONS];
+
+  rdim = GFC_DESCRIPTOR_EXTENT(shape,0);
+  if (rdim != GFC_DESCRIPTOR_RANK(ret))
+    runtime_error("rank of return array incorrect in RESHAPE intrinsic");
+
+  shape_empty = 0;
+
+  for (n = 0; n < rdim; n++)
+    {
+      shape_data[n] = shape->data[n * GFC_DESCRIPTOR_STRIDE(shape,0)];
+      if (shape_data[n] <= 0)
+	{
+	  shape_data[n] = 0;
+	  shape_empty = 1;
+	}
+    }
 
   if (ret->data == NULL)
     {
-      rdim = shape->dim[0].ubound - shape->dim[0].lbound + 1;
       rs = 1;
       for (n = 0; n < rdim; n++)
 	{
-	  ret->dim[n].lbound = 0;
-	  rex = shape->data[n * shape->dim[0].stride];
-	  ret->dim[n].ubound =  rex - 1;
-	  ret->dim[n].stride = rs;
+	  rex = shape_data[n];
+
+	  GFC_DIMENSION_SET(ret->dim[n],0,rex - 1,rs);
+
 	  rs *= rex;
 	}
       ret->offset = 0;
       ret->data = internal_malloc_size ( rs * size );
       ret->dtype = (source->dtype & ~GFC_DTYPE_RANK_MASK) | rdim;
     }
-  else
-    {
-      rdim = GFC_DESCRIPTOR_RANK (ret);
-    }
 
-  rsize = 1;
-  for (n = 0; n < rdim; n++)
-    {
-      if (order)
-        dim = order->data[n * order->dim[0].stride] - 1;
-      else
-        dim = n;
-
-      rcount[n] = 0;
-      rstride[n] = ret->dim[dim].stride;
-      rextent[n] = ret->dim[dim].ubound + 1 - ret->dim[dim].lbound;
-
-      if (rextent[n] != shape->data[dim * shape->dim[0].stride])
-        runtime_error ("shape and target do not conform");
-
-      if (rsize == rstride[n])
-        rsize *= rextent[n];
-      else
-        rsize = 0;
-      if (rextent[n] <= 0)
-        return;
-    }
-
-  sdim = GFC_DESCRIPTOR_RANK (source);
-  ssize = 1;
-  sempty = 0;
-  for (n = 0; n < sdim; n++)
-    {
-      scount[n] = 0;
-      sstride[n] = source->dim[n].stride;
-      sextent[n] = source->dim[n].ubound + 1 - source->dim[n].lbound;
-      if (sextent[n] <= 0)
-	{
-	  sempty = 1;
-	  sextent[n] = 0;
-	}
-
-      if (ssize == sstride[n])
-        ssize *= sextent[n];
-      else
-        ssize = 0;
-    }
+  if (shape_empty)
+    return;
 
   if (pad)
     {
@@ -143,8 +110,8 @@ reshape_internal (parray *ret, parray *source, shape_type *shape,
       for (n = 0; n < pdim; n++)
         {
           pcount[n] = 0;
-          pstride[n] = pad->dim[n].stride;
-          pextent[n] = pad->dim[n].ubound + 1 - pad->dim[n].lbound;
+          pstride[n] = GFC_DESCRIPTOR_STRIDE(pad,n);
+          pextent[n] = GFC_DESCRIPTOR_EXTENT(pad,n);
           if (pextent[n] <= 0)
 	    {
 	      pempty = 1;
@@ -166,6 +133,104 @@ reshape_internal (parray *ret, parray *source, shape_type *shape,
       pptr = NULL;
     }
 
+  if (unlikely (compile_options.bounds_check))
+    {
+      index_type ret_extent, source_extent;
+
+      rs = 1;
+      for (n = 0; n < rdim; n++)
+	{
+	  rs *= shape_data[n];
+	  ret_extent = GFC_DESCRIPTOR_EXTENT(ret,n);
+	  if (ret_extent != shape_data[n])
+	    runtime_error("Incorrect extent in return value of RESHAPE"
+			  " intrinsic in dimension %ld: is %ld,"
+			  " should be %ld", (long int) n+1,
+			  (long int) ret_extent, (long int) shape_data[n]);
+	}
+
+      source_extent = 1;
+      sdim = GFC_DESCRIPTOR_RANK (source);
+      for (n = 0; n < sdim; n++)
+	{
+	  index_type se;
+	  se = GFC_DESCRIPTOR_EXTENT(source,n);
+	  source_extent *= se > 0 ? se : 0;
+	}
+
+      if (rs > source_extent && (!pad || pempty))
+	runtime_error("Incorrect size in SOURCE argument to RESHAPE"
+		      " intrinsic: is %ld, should be %ld",
+		      (long int) source_extent, (long int) rs);
+
+      if (order)
+	{
+	  int seen[GFC_MAX_DIMENSIONS];
+	  index_type v;
+
+	  for (n = 0; n < rdim; n++)
+	    seen[n] = 0;
+
+	  for (n = 0; n < rdim; n++)
+	    {
+	      v = order->data[n * GFC_DESCRIPTOR_STRIDE(order,0)] - 1;
+
+	      if (v < 0 || v >= rdim)
+		runtime_error("Value %ld out of range in ORDER argument"
+			      " to RESHAPE intrinsic", (long int) v + 1);
+
+	      if (seen[v] != 0)
+		runtime_error("Duplicate value %ld in ORDER argument to"
+			      " RESHAPE intrinsic", (long int) v + 1);
+		
+	      seen[v] = 1;
+	    }
+	}
+    }
+
+  rsize = 1;
+  for (n = 0; n < rdim; n++)
+    {
+      if (order)
+        dim = order->data[n * GFC_DESCRIPTOR_STRIDE(order,0)] - 1;
+      else
+        dim = n;
+
+      rcount[n] = 0;
+      rstride[n] = GFC_DESCRIPTOR_STRIDE(ret,dim);
+      rextent[n] = GFC_DESCRIPTOR_EXTENT(ret,dim);
+
+      if (rextent[n] != shape_data[dim])
+        runtime_error ("shape and target do not conform");
+
+      if (rsize == rstride[n])
+        rsize *= rextent[n];
+      else
+        rsize = 0;
+      if (rextent[n] <= 0)
+        return;
+    }
+
+  sdim = GFC_DESCRIPTOR_RANK (source);
+  ssize = 1;
+  sempty = 0;
+  for (n = 0; n < sdim; n++)
+    {
+      scount[n] = 0;
+      sstride[n] = GFC_DESCRIPTOR_STRIDE(source,n);
+      sextent[n] = GFC_DESCRIPTOR_EXTENT(source,n);
+      if (sextent[n] <= 0)
+	{
+	  sempty = 1;
+	  sextent[n] = 0;
+	}
+
+      if (ssize == sstride[n])
+        ssize *= sextent[n];
+      else
+        ssize = 0;
+    }
+
   if (rsize != 0 && ssize != 0 && psize != 0)
     {
       rsize *= size;
@@ -185,16 +250,16 @@ reshape_internal (parray *ret, parray *source, shape_type *shape,
 
   if (sempty)
     {
-      /* Switch immediately to the pad array.  */
+      /* Pretend we are using the pad array the first time around, too.  */
       src = pptr;
-      sptr = NULL;
+      sptr = pptr;
       sdim = pdim;
       for (dim = 0; dim < pdim; dim++)
 	{
 	  scount[dim] = pcount[dim];
 	  sextent[dim] = pextent[dim];
 	  sstride[dim] = pstride[dim];
-	  sstride0 = sstride[0] * size;
+	  sstride0 = pstride[0] * size;
 	}
     }
 
@@ -282,16 +347,33 @@ reshape (parray *ret, parray *source, shape_type *shape, parray *pad,
 		    GFC_DESCRIPTOR_SIZE (source));
 }
 
-extern void reshape_char (parray *, GFC_INTEGER_4, parray *, shape_type *,
-			  parray *, shape_type *, GFC_INTEGER_4,
-			  GFC_INTEGER_4);
+
+extern void reshape_char (parray *, gfc_charlen_type, parray *, shape_type *,
+			  parray *, shape_type *, gfc_charlen_type,
+			  gfc_charlen_type);
 export_proto(reshape_char);
 
 void
-reshape_char (parray *ret, GFC_INTEGER_4 ret_length __attribute__((unused)),
+reshape_char (parray *ret, gfc_charlen_type ret_length __attribute__((unused)),
 	      parray *source, shape_type *shape, parray *pad,
-	      shape_type *order, GFC_INTEGER_4 source_length,
-	      GFC_INTEGER_4 pad_length __attribute__((unused)))
+	      shape_type *order, gfc_charlen_type source_length,
+	      gfc_charlen_type pad_length __attribute__((unused)))
 {
   reshape_internal (ret, source, shape, pad, order, source_length);
 }
+
+
+extern void reshape_char4 (parray *, gfc_charlen_type, parray *, shape_type *,
+			   parray *, shape_type *, gfc_charlen_type,
+			   gfc_charlen_type);
+export_proto(reshape_char4);
+
+void
+reshape_char4 (parray *ret, gfc_charlen_type ret_length __attribute__((unused)),
+	       parray *source, shape_type *shape, parray *pad,
+	       shape_type *order, gfc_charlen_type source_length,
+	       gfc_charlen_type pad_length __attribute__((unused)))
+{
+  reshape_internal (ret, source, shape, pad, order,
+		    source_length * sizeof (gfc_char4_t));
+}
